/* ucla_kernel.cu CUDA kernel for the UCLA model */

////////////////////////////////////////////////////////////////////////////////
// Macros and definitions
////////////////////////////////////////////////////////////////////////////////

//#include <cmath>
//#include <cfloat>


#ifndef NULL
 #define NULL 0
#endif
//#define T_END  1200.0 //600.0
//#define H_EULER  0.001 //0.1
//#define DT_STIM  1.0
//#define I_STIM  40.0
//#define DT_OUT   1.0
//#define EULER_STIM_MULT 10//10
//#define FILE_NAME "ucla_ap.txt"


#define NDIM 31//28
  
// Define standard floating point type.
typedef FLOAT Real;
// Math macros. Adapt these according to definition of Real.
#define ABS abs//fabsf
#define SQRT sqrt//sqrtf
#define EXPH exp//expf
#define SQR(A) ((A)*(A))
#define LOGH log//logf
#define RCONST(x) x##F
#define RPowerI pow//powf
#define MIN_DIFF RCONST(0.001)

// Euler settings in global_defs


/* Independent parameters. */
#define xnao        RCONST(136.0)
#define xki         RCONST(140.0)
#define xko         RCONST(5.4)
#define Ecao         RCONST(1.8)

#define Egca         RCONST(182.0)
#define ica_factor  RCONST(1.0)
#define gtos        RCONST(0.04/2.0)
#define gtof        RCONST(0.11/2.0)

#define gnaca       RCONST(0.84)
#define gkr         RCONST(0.0125/2.0)
#define gks         RCONST(0.32/2.0)
#define gk1         RCONST(0.3/2.0)
#define gnak        RCONST(1.5)
#define Evup         RCONST(0.4)
#define taus        RCONST(4.0)
#define gna         RCONST(12.0)
#define taur        RCONST(30.0)
#define taua        RCONST(100.0)
#define av          RCONST(11.3)
#define cstar       RCONST(90.0)
#define kmnsca      RCONST(0.45)
#define ginsca      RCONST(0.0)

/* Constants */
#define EnnoT   RCONST(308.0)
#define Vc  RCONST(-80.0)
#define xxr RCONST(8.314)
#define xf  RCONST(96.485)
#define frt (xf/(xxr*EnnoT))

/* Aliases of state variables. */
#define Ev     _STATE_X[0]
//#define xm    _STATE_X[1]
//#define xh    _STATE_X[2]
//#define xj    _STATE_X[3]
//#define xr    _STATE_X[4]
//#define xs1   _STATE_X[5]
//#define xs2   _STATE_X[6]
//#define cp    _STATE_X[7]
//#define cs    _STATE_X[8]
//#define ci    _STATE_X[9]
//#define cj    _STATE_X[10]
//#define cjp   _STATE_X[11]
#define xir   _STATE_X[12] // Jrel
/* Markov gate variables */
#define c1    _STATE_X[13]
#define c2    _STATE_X[14]
#define xi1ca _STATE_X[15]
#define xi1ba _STATE_X[16]
#define xi2ca _STATE_X[17]
#define xi2ba _STATE_X[18]
#define xnai  _STATE_X[19]
#define xtof  _STATE_X[20]
#define ytof  _STATE_X[21]
#define xtos  _STATE_X[22]
#define ytos  _STATE_X[23]
#define rtos  _STATE_X[24]
#define tropi _STATE_X[25]
#define trops _STATE_X[26]


////////////////////////////////////////////////////////////////////////////////
// Local function declarations
////////////////////////////////////////////////////////////////////////////////
int ucla_euler_step(const Real h, Real *_STATE_X, Real In, Real *parameters);
Real comp_ina(Real h, Real *_STATE_X, Real& dxh, Real& dxj, Real& dxm, Real *parameters);
Real comp_ikr(Real h, Real *_STATE_X, Real& dxr, Real *parameters);
Real comp_iks(Real h, Real *_STATE_X, Real& dxs1, Real& dxs2, Real *parameters);
Real comp_ik1 (Real *_STATE_X, Real *parameters);
Real comp_ito(Real dtdt, Real *_STATE_X, Real& dxtos, Real& dytos, Real& drtos, Real& dxtof, Real& dytof, Real *parameters);
Real comp_inak(Real *_STATE_X, Real *parameters);
Real comp_inaca(Real *_STATE_X, Real *parameters);
Real comp_icalpo(Real dtdt, Real *_STATE_X, Real& dc1, Real& dc2, Real& dxi1ca, Real& dxi2ca, Real& dxi1ba, Real& dxi2ba, Real *parameters);
Real comp_iup(Real *_STATE_X, Real *parameters);
Real comp_ileak(Real *_STATE_X);
Real comp_inst_buffer(Real c, Real *parameters);
Real comp_rxa (Real *_STATE_X, Real *parameters);
Real comp_Q(Real *_STATE_X, Real *parameters);
Real comp_dir(Real po, Real Qr, Real rxa, Real *_STATE_X, Real dcj, Real *parameters);
Real comp_dcp (Real po, Real Qr, Real rxa, Real *_STATE_X, Real *parameters);
Real comp_insca(Real *_STATE_X, Real *parameters);

////////////////////////////////////////////////////////////////////////////////
//! Compute an adaptive Euler step of the LR1 model using the Rush&Larsen 
//  method and a simple time step adaptation based on the value of dv/dt.
//  *Input*
//  h        - step size in ms.
//  _STATE_X - NDIM-element state vector with state at current time t.
//  In       - Stimulation current in uA/uF. Defined as an inward positive current!
//  *Output*
//  _STATE_X - Contains the state at time t+h.
////////////////////////////////////////////////////////////////////////////////
int ucla_one_step(const Real h, Real *x, Real In, Real *parameters) {
  // Save old state
  Real x_old[NDIM];
  for (unsigned int i = 0; i < NDIM; i++)
    x_old[i] = x[i];
  
  // Try a single Euler step
  ucla_euler_step(h, x, In, parameters);
  
  // Time step adaptation
  if (ABS(x_old[0]-x[0]) > DVDT_MAX) {
    // Take TS_ADAPT_FCT steps of size dt/TS_ADAPT_FCT, using x_cell_old as
    // temporary result buffer.
    for (unsigned int i = 0; i < TS_ADAPT_FCT; i++)
      ucla_euler_step(h/TS_ADAPT_FCT, x_old, In, parameters);
    // Report result back to x_cell.
    for (unsigned int i = 0; i < NDIM; i++)
      x[i] = x_old[i];        
  }
  
  return(0);
}

int ucla_euler_step(const Real h, Real *_STATE_X, Real In, Real *parameters)
{
	Real xik1, xito, xinak, xinacaq, xdif, xiup, xileak;
	Real po, rxa, xicaq, Qr, xina, xikr;
	Real xiks;
        #ifndef USE_CA_CELL 
           Real xinaca,xica;
        #endif
  Real insca;
  // Derivatives
  Real dxm, dxh, dxj, dxr, dxs1, dxs2, dcp, dcs, dci, dcj, dcjp;
  Real dxir, dc1, dc2, dxi1ca, dxi1ba, dxi2ca, dxi2ba, dxnai, dxtof;
  Real dytof, dxtos, dytos, drtos, dtropi, dtrops;

  const Real xkon=0.0327f;
  const Real xkoff=0.0196f;
  const Real btrop=70.0f;
  //const Real wca=8.0313f; //conversion factor between micro molar/ms to micro amps/ micro farads
	const Real xrr=(1.0f/wca)/1000.0f; // note: sodium is in m molar so need to divide by 
  
  //Real v_old = Ev; // For AP onset detection.

	// -------- Compute ion channel currents.
  // Sodium current
  xina = comp_ina(h, _STATE_X, dxh, dxj, dxm, parameters);
  // Inward rectifier K+ current IK1.
  xik1 = comp_ik1(_STATE_X, parameters);
  // Inward K+ current Ito
  xito = comp_ito(h, _STATE_X, dxtos, dytos, drtos, dxtof, dytof, parameters);
  // Delayed rectifier K+ currents (rapid and slow).
  xikr = comp_ikr(h, _STATE_X, dxr, parameters);
  xiks = comp_iks(h, _STATE_X, dxs1, dxs2, parameters);
  
  // --------- Compute pump currents.
  // Na-K pump.
  xinak = comp_inak(_STATE_X, parameters);
  // Na-Ca exchange ion flux.
  xinacaq = comp_inaca(_STATE_X, parameters);
  #ifndef USE_CA_CELL 
   xinaca = wca*xinacaq; // Convert flux to current.
  #endif
  // --------- Compute Ca cycling.
  xdif = (_STATE_X[8]-_STATE_X[9])/taus; //diffusion from submembrane to myoplasm
  // Nonlinear buffering (Troponin).
  dtropi = xkon*_STATE_X[9]*(btrop - tropi) - xkoff*tropi;
  dtrops = xkon*_STATE_X[8]*(btrop - trops) - xkoff*trops;

  xiup = comp_iup(_STATE_X, parameters);
  xileak = comp_ileak(_STATE_X);

  po = comp_icalpo(h, _STATE_X, dc1, dc2, dxi1ca, dxi2ca, dxi1ba, dxi2ba, parameters);
  _STATE_X[30]=po;// mike UCLA
  rxa = comp_rxa(_STATE_X, parameters); // iCa in the paper.

  xicaq = Egca*po*rxa;// Ca current in micro M/ms

  // Derivatives of Ca gating variables.
  dcs = comp_inst_buffer(_STATE_X[8],parameters)* \
          (50.0f*(xir-xdif-xicaq+xinacaq)-dtrops);
  dci = comp_inst_buffer(_STATE_X[9],parameters)*(xdif - xiup + xileak - dtropi);
  dcj = -xir + xiup - xileak; // SR load dynamics
  dcjp = (_STATE_X[10] - _STATE_X[11])/taua; // NSR-JSR relaxation dynamics
  Qr = comp_Q(_STATE_X, parameters);
  dxir = comp_dir(po, Qr, rxa, _STATE_X, dcj, parameters); // dJrel/dt
  dcp = comp_dcp(po, Qr, rxa, _STATE_X, parameters);
 
  // Total Ca current.
  #ifndef USE_CA_CELL 
   xica = 2.0f*wca*xicaq;
  #endif
  // Addition in Sato et al. PNAS: Non specific Ca-activated current from LRd.
  // Only present in models 3 and 4.
  insca = ginsca*comp_insca(_STATE_X, parameters);   
  
 // sodium dynamics
  #ifdef USE_CA_CELL 
   dxnai = -xrr*(xina + 3.0f*xinak + 3.0f*_STATE_X[27]);
  #else
   dxnai = -xrr*(xina + 3.0f*xinak + 3.0f*xinaca);
  #endif  
  // Compute derivatives.
  //Ev += h*(-(xina+xik1+xikr+xiks+xito+xinaca+xinak+xica+insca) + In);
  
  #ifdef USE_CA_CELL 
   Ev += h*(-(xina+xik1+xikr+xiks+xito+_STATE_X[27]+xinak+_STATE_X[28]+insca) + In); 
   //Ev += h*(-(xina+xik1+xikr+xiks+xito+_STATE_X[27]+xinak+xica+insca) + In); _STATE_X[28]=xica;
  #else  
   Ev += h*(-(xina+xik1+xikr+xiks+xito+xinaca+xinak+xica+insca) + In);
   _STATE_X[27]=xinaca;
   _STATE_X[28]=xica;
  #endif

  _STATE_X[1] += dxm*h;
  _STATE_X[2] += dxh*h; 
  _STATE_X[3]    += dxj*h;
  _STATE_X[4]    += dxr*h;
  _STATE_X[5]   += dxs1*h;
  _STATE_X[6]   += dxs2*h;
  _STATE_X[7]    += dcp*h;
  _STATE_X[8]    += dcs*h;
  _STATE_X[9]    += dci*h;	//???????? why still using this if we use whole cell? 
  _STATE_X[10]    += dcj*h;
  _STATE_X[11]   += dcjp*h;
  xir   += dxir*h; // Jrel
  c1    += dc1*h;
  c2    += dc2*h;
  xi1ca += dxi1ca*h;
  xi1ba += dxi1ba*h;
  xi2ca += dxi2ca*h;
  xi2ba += dxi2ba*h;
#ifndef CLAMP_NAI
  xnai  += dxnai*h;
#endif
  xtof  += dxtof*h;
  ytof  += dytof*h;
  xtos  += dxtos*h;
  ytos  += dytos*h;
  rtos  += drtos*h;
  tropi += dtropi*h;
  trops += dtrops*h;
  
  return 0;
}


/* -------------------------------------------------------------
 * Model currents
 */
//-----------	sodium current following Hund-Rudy -------------------
Real comp_ina(Real dtdt, Real *_STATE_X, Real& dxh, Real& dxj, Real& dxm, 
                         Real *parameters)
{
	Real ena, a, am, bm, ah, bh, aj, bj;
  Real taum, tauj, tauh;
	Real xina;
  ena = (1.0/frt)*LOGH(xnao/xnai);

  // Newer formulation (Leonid Livschitz).
  a = 1.0 - 1.0/(1.0+EXPH(-(Ev+40.0)/0.24));
  ah = a*0.135*EXPH((RCONST(80.0)+Ev)/(-6.8));
  bh = (1.0-a)/(0.13*(1+EXPH((Ev+10.66)/(-11.1)))) + \
        a*(3.56*EXPH(0.079*Ev)+3.1*1.0e5*EXPH(0.35*Ev));
  aj =  a*(-1.2714e5*EXPH(0.2444*Ev)-3.474e-5*EXPH(-0.04391*Ev))*(Ev+37.78)/(1.0+EXPH(0.311*(Ev+79.23)));
  bj = (1.0-a)*(0.3*EXPH(-2.535e-7*Ev)/(1+EXPH(-0.1*(Ev+32)))) + \
        a*(0.1212*EXPH(-0.01052*Ev)/(1+EXPH(-0.1378*(Ev+40.14))));

  if (ABS(Ev+47.13) < 0.01)
    am = 3.2;
  else
    am = 0.32*(Ev+47.13)/(1.0 - EXPH(-0.1*(Ev+47.13)));
  bm = 0.08*EXPH(-Ev/11.0);    
    
  xina= gna*_STATE_X[2]*_STATE_X[3]*_STATE_X[1]*_STATE_X[1]*_STATE_X[1]*(Ev - ena);  

  // Rush-Larsen method.
  tauh = 1.0/(ah+bh);
	tauj = 1.0/(aj+bj);
	taum = 1.0/(am+bm); 
	dxh = (ah*tauh - (ah*tauh - _STATE_X[2])*EXPH(-dtdt/tauh) - _STATE_X[2])/dtdt;
	dxj = (aj*tauj - (aj*tauj - _STATE_X[3])*EXPH(-dtdt/tauj) - _STATE_X[3])/dtdt;
	dxm = (am*taum - (am*taum - _STATE_X[1])*EXPH(-dtdt/taum) - _STATE_X[1])/dtdt; 
    
  return xina;
}

//-------------- Ikr following Shannon------------------
Real comp_ikr(Real dtdt, Real *_STATE_X, Real& dxr, Real *parameters)
{
  const Real ek = (1.0/frt)*LOGH (xko/xki);// K reversal potential
  const Real gss = SQRT(xko/5.4);
  Real xkrv1, xkrv2, taukr, xkrinf, rg, xikr;
  
  if (ABS(Ev + 7.0) < 0.001/0.123)
    xkrv1=0.00138/0.123;
  else
    xkrv1=0.00138*(Ev + 7.0)/(1.0 - EXPH(-0.123*(Ev + 7.0)));
	
  if (ABS(Ev + 10.0) < 0.001/0.145 )
    xkrv2=0.00061/0.145;
  else
    xkrv2=0.00061*(Ev+10.0)/(EXPH(0.145*(Ev+10.0)) - 1.0);
  
	taukr = 1.0/(xkrv1 + xkrv2);
  xkrinf = 1.0/(1.0 + EXPH(-(Ev + 50.0)/7.5));
  rg = 1.0/(1.0 + EXPH((Ev + 33.0)/22.4));
  xikr = gkr*gss*_STATE_X[4]*rg*(Ev - ek);
  
  // Rush-Larsen method.
  dxr = (xkrinf - (xkrinf - _STATE_X[4])*EXPH(-dtdt/taukr) - _STATE_X[4])/dtdt;
  
	return xikr;
}

// ----- Iks modified from Shannon, with new Ca dependence------------
Real comp_iks(Real dtdt, Real *_STATE_X, Real& dxs1, Real& dxs2, 
                         Real *parameters)
{
  const Real prnak=0.018330;
	Real eks, xs1ss, xs2ss, tauxs1, tauxs2, gksx, xiks;
	
  eks = (1.0/frt)*LOGH((xko+prnak*xnao)/(xki+prnak*xnai));
  xs1ss = 1.0/(1.0 + EXPH(-(Ev-1.50)/16.70));
  xs2ss = xs1ss;
  
  // Adapted from original.
  if (ABS(Ev + 30.0) < 0.001/0.0687)
    tauxs1=1.0/(0.0000719/0.148+0.000131/0.0687);
  else
    tauxs1 = 1.0/(0.0000719*(Ev + 30.0)/(1.0 - EXPH(-0.148*(Ev + 30.0))) + 
                  0.000131*(Ev + 30.0)/(EXPH(0.0687*(Ev + 30.0)) - 1.0));
  tauxs2 = 4.0*tauxs1;
  //gksx = 0.433*(1.0 + 0.8/(1.0 + RPowerI((Real)(0.5/_STATE_X[9]),3))); // (q_Ks) 
  
  #ifdef USE_CA_CELL 
   gksx = 0.433*(1.0 + 0.8/(1.0 + RPowerI((Real)(0.5/_STATE_X[29]),3)));
  #else
   gksx = 0.433*(1.0 + 0.8/(1.0 + RPowerI((Real)(0.5/_STATE_X[9]),3)));
  #endif 

  xiks = gks*gksx*_STATE_X[5]*_STATE_X[6]*(Ev-eks);
  
  // Rush-Larsen method.
  dxs1 = (xs1ss - (xs1ss - _STATE_X[5])*EXPH(-dtdt/tauxs1) - _STATE_X[5])/dtdt;
  dxs2 = (xs2ss - (xs2ss - _STATE_X[6])*EXPH(-dtdt/tauxs2) - _STATE_X[6])/dtdt;
    
	return xiks;
}

//------Ik1 following Luo-Rudy formulation (from Shannon model) ------
Real comp_ik1 ( Real *_STATE_X, Real *parameters )
{
	Real ek, aki, bki, xkin, xik1;
  const Real gki = SQRT(xko/5.4);
  
	ek = (1.0/frt)*LOGH(xko/xki); // K reversal potential
  aki =1.02/ ( 1.0+EXPH ( 0.2385* ( Ev-ek-59.215 ) ) );
  bki = (0.49124*EXPH(0.08032*(Ev - ek + 5.476)) +
         EXPH(0.061750*(Ev - ek - 594.31)))/(1.0 + EXPH(-0.5143*(Ev - ek + 4.753)));
  xkin = aki/(aki + bki);
  xik1 = gk1*gki*xkin* (Ev - ek);
  return xik1;
}

//------- Ito fast following Shannon et. al. 2005 -----------
//------- Ito slow following Shannon et. al. 2005 -----------
Real comp_ito(Real dtdt, Real *_STATE_X, Real& dxtos, Real& dytos, 
                          Real& drtos, Real& dxtof, Real& dytof, Real *parameters)
{
  // -- Ito,s
  Real ek = 1.0/frt*LOGH(xko/xki);// K reversal potential
  Real rt1 = -(Ev + 3.0)/15.0;
  Real rt2 = (Ev + 33.5)/10.0;
  Real rt3 = (Ev + 60.0)/10.0;
  Real xtos_inf = 1.0/(1.0 + EXPH(rt1));
  Real ytos_inf = 1.0/(1.0 + EXPH(rt2));
  Real rs_inf = 1.0/(1.0 + EXPH(rt2));
  Real trs = (2800.0 - 500.0)/(1.0 + EXPH(rt3)) + 220.0 + 500.0;
  Real txs = 9.0/(1.0+EXPH(-rt1)) + 0.5;
  Real tys = 3000.0/(1.0+EXPH(rt3)) + 30.0;
  Real xitos = gtos*xtos*(0.5*ytos + 0.5*rtos)*(Ev - ek); // ito slow (Sato)
  
  // -- Ito,f
  Real xtof_inf = xtos_inf;
  Real ytof_inf = ytos_inf;
  Real txf = 3.5*EXPH(-(Ev/30.0)*(Ev/30.0)) + 1.5;
  Real tyf = 20.0/(1.0+EXPH(rt2)) + 20.0;
  Real xitof = gtof*xtof*ytof*(Ev - ek); // ito fast

  // Rush-Larsen method.
  dxtos = (xtos_inf - (xtos_inf-xtos)*EXPH(-dtdt/txs) - xtos)/dtdt;
  dytos = (ytos_inf - (ytos_inf-ytos)*EXPH(-dtdt/tys) - ytos)/dtdt;
  drtos = (rs_inf - (rs_inf-rtos)*EXPH(-dtdt/trs) - rtos)/dtdt;
  dxtof = (xtof_inf - (xtof_inf-xtof)*EXPH(-dtdt/txf) - xtof)/dtdt;
  dytof = (ytof_inf - (ytof_inf-ytof)*EXPH(-dtdt/tyf) - ytof)/dtdt;
    
	return xitos + xitof;
}

// -------Inak (sodium-potassium exchanger) following Shannon --------------
Real comp_inak ( Real *_STATE_X, Real *parameters )
{
	Real fnak, xinak;
  const Real xkmko=1.5;	//these are Inak constants adjusted to fit
  //the EXPerimentally measured dynamic restitution curve
  const Real xkmnai=12.0;
  const Real sigma = ( EXPH ( xnao/67.3 )-1.0 ) /7.0;
	
  fnak = 1.0/ ( 1+0.1245*EXPH ( -0.1*Ev*frt ) +0.0365*sigma*EXPH ( -Ev*frt ) );
  xinak = gnak*fnak* ( 1./ ( 1.+ ( xkmnai/xnai ) ) ) *xko/ ( xko+xkmko );

	return xinak;
}

// --- Inaca (sodium-calcium exchange) following Shannon and Hund-Rudy------
//	Note: all concentrations are in mM
Real comp_inaca ( Real *_STATE_X, Real *parameters )
{
	Real zw3, zw4, aloss, yz1, yz2, yz3, yz4, zw8, xinacaq, csm;
  const Real xkdna=0.3; // (cnaca)
  const Real xmcao=1.3;
  const Real xmnao=87.5;
  const Real xmnai=12.3;
  const Real xmcai=0.0036;
  csm = _STATE_X[8]/1000.0; // cs is in uMol, but these eqns are in mMol
	
  zw3 = RPowerI(xnai,3)*Ecao*EXPH(Ev*0.35*frt) - RPowerI(xnao,3)*csm*EXPH(Ev*(0.35-1.)*frt);
  zw4 = 1.0+0.2*EXPH ( Ev* ( 0.35-1.0 ) *frt );
  aloss = 1.0/(1.0+RPowerI((xkdna/_STATE_X[8]),3));
  yz1 = xmcao*RPowerI ( xnai,3 ) +RPowerI ( xmnao,3 ) *csm;
  yz2 = RPowerI ( xmnai,3 ) *Ecao* ( 1.0+csm/xmcai );
  yz3 = xmcai*RPowerI ( xnao,3 ) * ( 1.0+RPowerI ( ( xnai/xmnai ),3 ) );
  yz4 = RPowerI ( xnai,3 ) *Ecao+RPowerI ( xnao,3 ) *csm;
  zw8 = yz1+yz2+yz3+yz4;
	
  xinacaq = gnaca*aloss*zw3/ ( zw4*zw8 );

	return xinacaq;
}


//	compute driving force (iCa)
Real comp_rxa(Real *_STATE_X, Real *parameters)
{
	Real za, factor, rxa, csm;
  Real factor1;
  const Real pca = 0.00054;
  csm = _STATE_X[8]/1000.0;
  
	za = Ev*2.0*frt;
  factor1 = 4.0*pca*xf*xf/(xxr*EnnoT); // Temperature is a factor here!
  factor = Ev*factor1;
  if (ABS(za) < 0.001)
    rxa = factor1*(csm*EXPH(za) - 0.341*(Ecao))/(2.0*frt);
  else
    rxa = factor*(csm*EXPH(za) - 0.341*(Ecao))/(EXPH(za) - 1.0);
	return rxa;
}

// ------ Markovian Ca current --------------------------------
//	Markov model:All parameters have been fitted directly to
//	experimental current traces using a multidimensional current fitting
//	routine.
Real comp_icalpo(Real dtdt, Real *_STATE_X, Real& dc1, Real& dc2, 
                            Real& dxi1ca, Real& dxi2ca, Real& dxi1ba, 
                            Real& dxi2ba, Real *parameters)
{
  Real P_o, P_o_inf, P_r, P_s, P_tmp;
  Real alpha, beta, f_c_p, R_V, T_Ca, tau_Ca, tau_Ba;
	Real s1, s2, s2_p, k1;
	Real k3, k3_p, k6, k5, k6_p, k5_p, k4, k4_p;
	
  // Original parameter values from table 4
  const Real cp_tilde = 3.0;   // 
  const Real cp_bar = 6.09365; // Does not appear in table 4; Change by YFX
  const Real tau_po=1.0;
  const Real r1 = 0.3;
  const Real r2 = 3.0;
  const Real s1_p=0.00195;
  const Real k1_p=0.00413;
  const Real k2k2 = 1.03615e-4;
  const Real k2_p = 0.00224;
  const Real T_Ba = 450.0;
  
	P_o_inf = 1.0/(1.0 + EXPH(-Ev/8.0));
  P_r = 1.0 - 1.0/(1.0+EXPH(-(Ev+40.0)/4.0)); // Erratum in Mahajan et al.?
  P_s = 1.0/(1.0+EXPH(-(Ev+40.0)/11.32));
  P_tmp = EXPH(-(Ev+40.0)/3.0);    // For readability (not in original publication)
  
  alpha = P_o_inf/tau_po;
  // MOD EDL2010MAY03: Avoid potential division by zero below by taking BETA
  // larger than 0.0f  

  beta = max(1.0e-6,(1.0 - P_o_inf)/tau_po);
  
  f_c_p = 1.0/(1.0 + RPowerI(((Real)(cp_tilde/_STATE_X[7])),3)); // f(c_p)

  R_V = 10.0 + 4954.0*EXPH(Ev/15.6);
  T_Ca = (78.0329 + 0.1*(1.0 + RPowerI(((Real)(_STATE_X[7]/cp_bar)),4)))/(1.0 + RPowerI(((Real)(_STATE_X[7]/cp_bar)),4));
  
  tau_Ca = (R_V - T_Ca)*P_r + T_Ca;
  tau_Ba = (R_V - T_Ba)*P_r + T_Ba;

  s1 = 0.0182688*f_c_p;  // Change by YFX for DAD model.
  k1 = 0.024168*f_c_p;   // Change by YFX for DAD model.
  s2 = s1*(k2k2/k1)*(r1/r2);
  s2_p =  s1_p*(k2_p/k1_p)*(r1/r2);
  
  k3 = P_tmp/(3.0*(1.0+P_tmp));
  k3_p = k3;
  k5 = (1.0f - P_s)/tau_Ca;
  k5_p = (1.0f - P_s)/tau_Ba;
  k6 = f_c_p*P_s/tau_Ca;
  k6_p = P_s/tau_Ba;
  k4 = k3*(alpha/beta)*(k1/k2k2)*(k5/k6);
  k4_p = k3_p*(alpha/beta)*(k1_p/k2_p)*(k5_p/k6_p);
  
  P_o = 1.0-xi1ca-xi2ca-xi1ba-xi2ba-c1-c2;
  
	// Derivatives of Markovian state variables
  dc2 = beta*c1 + k5*xi2ca + k5_p*xi2ba - (k6+k6_p+alpha)*c2;
  dc1 = alpha*c2 + k2k2*xi1ca + k2_p*xi1ba + r2*P_o - (beta+r1+k1_p+k1)*c1;
  dxi1ca = k1*c1 + k4*xi2ca + s1*P_o - (k3+k2k2+s2)*xi1ca;
  dxi2ca = k3*xi1ca + k6*c2 - (k5+k4)*xi2ca;
  dxi1ba = k1_p*c1 + k4_p*xi2ba + s1_p*P_o - (k3_p+k2_p+s2_p)*xi1ba;
  dxi2ba = k3_p*xi1ba + k6_p*c2 - (k5_p+k4_p)*xi2ba;
	
  return P_o;
}


//----- SERCA2a uptake current ------------------------------------
Real comp_iup(Real *_STATE_X, Real *parameters)
{
  const Real xup = 0.5;// uptake threshold
	
  return Evup*_STATE_X[9]*_STATE_X[9]/ ( _STATE_X[9]*_STATE_X[9] + xup*xup );
}

// ---------leak from the SR--------------------------
Real comp_ileak(Real *_STATE_X)
{
  const Real Egleak = 0.00002069;
  const Real kj = 50.0;
  return Egleak * _STATE_X[10]*_STATE_X[10]/(_STATE_X[10]*_STATE_X[10]+kj*kj) * (_STATE_X[10]*16.667-_STATE_X[9]); //vsr/vcell=0.06
}

// ---------- buffer dynamics in the myoplasm -----------------------
//buffering to calmodulin and SR are instantaneous, while buffering to
//Troponin C is time dependent.These are important to have reasonable
//Ca transient.Note: we have buffering in the suB_membrane space and
//the myoplasm.
Real comp_inst_buffer(Real c, Real *parameters)
{
	Real SR_buf, Cal_buf, Mem_buf, Sarco_buf, ATP_buf;
  const Real B_Cd = 24.0;
  const Real K_Cd = 7.0;
  const Real B_SR = 47.0;
  const Real K_SR = 0.6;
  const Real B_mem = 15.0;
  const Real K_mem = 0.3;
  const Real B_sar = 42.0;
  const Real K_sar = 13.0;
  const Real B_atp = 5000.0; // The ATP buffer has large influence on the
  const Real K_atp = 200.0;  // amplitude of the EADs.
  
  SR_buf = B_SR*K_SR/((K_SR+c)*(K_SR+c)); // SR binding sites buffering.
  Cal_buf = B_Cd*K_Cd/((K_Cd+c)*(K_Cd+c));  // Calmodulin Ca buffering.
  Mem_buf = B_mem*K_mem/((K_mem+c)*(K_mem+c));  // Membrane Ca buffering.
  Sarco_buf = B_sar*K_sar/((K_sar+c)*(K_sar+c));  // SR Ca buffering
  ATP_buf = B_atp*K_atp/((K_atp+c)*(K_atp+c));  // ATP Ca buffering (not in original)	
  
  return 1.0/(1.0 + SR_buf + Cal_buf + Mem_buf + Sarco_buf + ATP_buf);
}

// --------- release-load functional dependence ----------------
Real comp_Q(Real *_STATE_X, Real *parameters)
{
	Real bv;
	
  bv = (cstar-50.0) - av*cstar;
  if (_STATE_X[11] < 50)
    return 0.0;
  else if ((_STATE_X[11] > 50.0) && (_STATE_X[11] < cstar))
    return _STATE_X[11] - 50.0;
  else
    return av*_STATE_X[11] + bv;  // bv is called s in the original publication.
}
/*Real comp_dir(Real po, Real Qr, Real rxa, 
                         Real *_STATE_X, Real *_STATE_DX,
                         Real *parameters)*/
Real comp_dir(Real po, Real Qr, Real rxa, 
                         Real *_STATE_X, Real dcj,
                         Real *parameters)
{
	Real spark_rate, gRyRV;
  const Real ay = 0.05;
  const Real gRyR = 2.58079f;

	gRyRV = gRyR*EXPH(-ay*(Ev+30.0))/(1.0+EXPH(-ay*(Ev+30.0)));
  spark_rate=ica_factor*po*ABS(rxa)*gRyRV; // minus and without ABS?
	
  return spark_rate*_STATE_X[10]*Qr/cstar - xir*(1.0-taur*dcj/_STATE_X[10])/taur;
}

// ----------- dyadic junction dynamics ------------------------
Real comp_dcp(Real po, Real Qr, Real rxa,
                         Real *_STATE_X, Real *parameters)
{
	Real gSRV, JSRtld, JCatld;
	const Real gSRbar=26841.8f; // m mol/(cm C)
  const Real ax = 0.3576f; 
  const Real gCabar = 9000.0f;; // m mol/(cm C)
  const Real taups = 0.5;

	gSRV  = gSRbar*EXPH(-ax*(Ev+30.0))/(1.0+EXPH(-ax*(Ev+30.0)));
  JSRtld = gSRV*Qr*po*ABS(rxa); // minus and without ABS?
  JCatld = ica_factor*gCabar*po*ABS(rxa); // minus and without ABS?
  
  return JSRtld + JCatld - (_STATE_X[7]-_STATE_X[8])/taups;
}

// ----------- Non specific Ca-activated current from LRd -----------
Real comp_insca(Real *_STATE_X, Real *parameters )
{
	Real ibarnsna, ibarnsk, insna, insk;
  const Real EnnoR = 8314;		// Universal Gas Constant (J/kmol*K)
  const Real frdy = 96485;	// Faraday's Constant (C/mol)
  const Real zna = 1.0;	// Na valence
  const Real zk = 1.0;	// K valence
  const Real ganai = 0.75;		// Activity coefficient of Na
  const Real ganao = 0.75;		// Activity coefficient of Na
  const Real gaki = 0.75;		// Activity coefficient of K
  const Real gako = 0.75;		// Activity coefficient of K
  const Real pnsca = 0.000000175;  // Permiability of channel to Na and K (cm/s)
	
  // Max. Na current through NSCa channel (uA/uF)
  if (ABS(EXPH((Ev*frdy)/(EnnoR*EnnoT)) - 1.0) < RCONST(1.0e-3)) {
    ibarnsna = pnsca*zna*frdy*(ganai*xnai - ganao*xnao);
    ibarnsk  = pnsca*zk*frdy*(gaki*xki - gako*xko);
  }
  else {
    ibarnsna = pnsca*zna*zna*((Ev*frdy*frdy)/(EnnoR*EnnoT))* \
    ((ganai*xnai*EXPH((zna*Ev*frdy)/(EnnoR*EnnoT)) - ganao*xnao )/ \
     (EXPH((zna*Ev*frdy)/(EnnoR*EnnoT)) - 1.0)); // (uA/uF)
    ibarnsk = pnsca*zk*zk*((Ev*frdy*frdy)/(EnnoR*EnnoT))* \
    ((gaki*xki*EXPH((zk*Ev*frdy)/(EnnoR*EnnoT)) - gako*xko ) / \
     (EXPH((zk*Ev*frdy)/(EnnoR*EnnoT))-1.0)); // (uA/uF)
  }
  insna = ibarnsna/(1.0 + RPowerI(kmnsca/_STATE_X[9],3));
  insk = ibarnsk/(1.0 + RPowerI(kmnsca/_STATE_X[9],3));
	
    
#ifdef __DEBUG_NANS_GPU
  if (isnan(insna+insk)) {
    // If a NaN was detected in dump the currents etc.
    printf("ibarnsna = %E\n",ibarnsna);
    printf("ibarnsk = %E\n",ibarnsk);
    printf("|Ev| = %E\n",ABS(Ev));
  }
#endif     
  
  return insna+insk;
}


////////////////////////////////////////////////////////////////////////////////
//! Set the standard initial conditions (which are defined in macros above).
////////////////////////////////////////////////////////////////////////////////
int ucla_get_default_initial_state(Real *_STATE_X)
{  
  if (_STATE_X == 0) return(-1);
  

  Ev=Vucla; // voltage
  
  _STATE_X[1]=0.00106085;// sodium m-gate
  _STATE_X[2]=0.990866;// sodium h-gate
  _STATE_X[3]=0.994012;// soiumj-gate

  _STATE_X[4]=0.0069357;// ikr gate variable
  _STATE_X[5]=0.0253915;// iks gate variable
  _STATE_X[6]=0.07509;// iks gate varaible

  _STATE_X[7]=1.16565;// averaged dyadic space con.
  _STATE_X[8]=0.192715;// averaged suB_membrane conc.
  _STATE_X[9]=0.215889;// myoplasm conc.

  _STATE_X[10]=99.5819;// NSR load
  _STATE_X[11]=94.8962;// average JSR load

  xir=0.00487311;// SR current flux

  // Markov gate variables
  c1=1.80428e-05;// C1
  c2=0.983776;// C2
  xi1ca=0.000216594;// I1_Ca
  xi1ba=3.33501e-05;// I1_Ba
  xi2ca=0.000781916;// I2_Ca
  xi2ba=0.0151724;// I2_Ba

  xnai=naiIC;// intracellular Na conc.

  xtof=0.00362348;// ito fast activation
  ytof=0.995164;// ito fast inactivation
  xtos=0.00362501;// ito slow activation
  ytos=0.211573;// ito slow inactivation
  rtos=0.434248;// ito slow inactivation

  tropi=19.2272;// time dependent buffers in myplasm (troponin)
  trops=17.3258;// time dependent buffers in suB_membrane (troponin)
  
  return(0);
}

/**
 * Function to initialize the parameter array in the model
 * structure
 */
int ucla_get_default_parameters(Real *parameters)
{
  if (parameters == NULL) return(1);
  
  // If using free parameters, set the default values below.
  
  return(0);
}
